-------------------------------------------------------------------------------
--
-- Copyright 2021
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- Author : R vd Walle
-- Purpose:  
--   Core design for Lofar2 SDP station
-- Description:
--   Combines sdp nodes.
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, technology_lib, diag_lib, dp_lib, tech_jesd204b_lib, wpfb_lib, tech_pll_lib, nw_10gbe_lib, eth_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE wpfb_lib.wpfb_pkg.ALL;
USE work.sdp_pkg.ALL;
USE eth_lib.eth_pkg.ALL;


ENTITY sdp_station IS
  GENERIC (
    g_technology             : NATURAL := c_tech_arria10_e1sg;
    g_sim                    : BOOLEAN := FALSE; --Overridden by TB
    g_wpfb                   : t_wpfb  := c_sdp_wpfb_subbands;
    g_bsn_nof_clk_per_sync   : NATURAL := c_sdp_f_adc_MHz*10**6;  -- Default 200M, overide for short simulation 
    g_scope_selected_subband : NATURAL := 0;
    g_use_fsub               : BOOLEAN := TRUE;
    g_use_xsub               : BOOLEAN := TRUE;
    g_use_bf                 : BOOLEAN := TRUE;
    g_P_sq                   : NATURAL := 1
  );
  PORT (
    -- System
    mm_clk        : IN STD_LOGIC;
    mm_rst        : IN STD_LOGIC := '0';
    
    dp_pps        : IN STD_LOGIC;
    dp_rst        : IN STD_LOGIC;
    dp_clk        : IN STD_LOGIC;

    -- ID
    gn_id         : IN STD_LOGIC_VECTOR(c_sdp_W_gn_id-1 DOWNTO 0);
    this_bck_id   : IN STD_LOGIC_VECTOR(6-1 DOWNTO 0);
    this_chip_id  : IN STD_LOGIC_VECTOR(2-1 DOWNTO 0);

    -- Transceiver clocks
    SA_CLK        : IN    STD_LOGIC := '0'; -- Clock 10GbE front (qsfp) and ring lines

     -- back transceivers (Note: numbered from 0)
    JESD204B_SERIAL_DATA       : IN    STD_LOGIC_VECTOR(c_sdp_S_pn-1 downto 0); 
                                                  -- Connect to the BCK_RX pins in the top wrapper
    JESD204B_REFCLK            : IN    STD_LOGIC; -- Connect to BCK_REF_CLK pin in the top level wrapper
 
    -- jesd204b syncronization signals
    JESD204B_SYSREF            : IN    STD_LOGIC;
    JESD204B_SYNC_N            : OUT   STD_LOGIC_VECTOR(c_sdp_S_pn-1 DOWNTO 0);


    ----------------------------------------------
    -- UDP Offload
    ----------------------------------------------
    udp_tx_sosi_arr            : OUT t_dp_sosi_arr(c_eth_nof_udp_ports-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
    udp_tx_siso_arr            : IN  t_dp_siso_arr(c_eth_nof_udp_ports-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);  

    ----------------------------------------------
    -- 10 GbE 
    ----------------------------------------------
    reg_nw_10GbE_mac_mosi      : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_nw_10GbE_mac_miso      : OUT t_mem_miso := c_mem_miso_rst;

    reg_nw_10GbE_eth10g_mosi   : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_nw_10GbE_eth10g_miso   : OUT t_mem_miso := c_mem_miso_rst;

    ----------------------------------------------
    -- AIT 
    ----------------------------------------------
    -- JESD
    jesd204b_mosi              : IN  t_mem_mosi := c_mem_mosi_rst;
    jesd204b_miso              : OUT t_mem_miso := c_mem_miso_rst;

    -- JESD control
    jesd_ctrl_mosi             : IN  t_mem_mosi := c_mem_mosi_rst;
    jesd_ctrl_miso             : OUT t_mem_miso := c_mem_miso_rst;

    -- Shiftram (applies per-antenna delay)
    reg_dp_shiftram_mosi       : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_shiftram_miso       : OUT t_mem_miso := c_mem_miso_rst;

    -- bsn source
    reg_bsn_source_v2_mosi     : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_bsn_source_v2_miso     : OUT t_mem_miso := c_mem_miso_rst;

    -- bsn scheduler
    reg_bsn_scheduler_wg_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_bsn_scheduler_wg_miso  : OUT t_mem_miso := c_mem_miso_rst;

    -- WG
    reg_wg_mosi                : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_wg_miso                : OUT t_mem_miso := c_mem_miso_rst;
    ram_wg_mosi                : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_wg_miso                : OUT t_mem_miso := c_mem_miso_rst;

    -- BSN MONITOR
    reg_bsn_monitor_input_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_bsn_monitor_input_miso : OUT t_mem_miso := c_mem_miso_rst;

    -- Data buffer bsn
    ram_diag_data_buf_bsn_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_diag_data_buf_bsn_miso : OUT t_mem_miso := c_mem_miso_rst;
    reg_diag_data_buf_bsn_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_diag_data_buf_bsn_miso : OUT t_mem_miso := c_mem_miso_rst;

    -- Aduh statistics monitor
    reg_aduh_monitor_mosi      : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_aduh_monitor_miso      : OUT t_mem_miso := c_mem_miso_rst;

    ----------------------------------------------
    -- FSUB 
    ----------------------------------------------
    -- Subband statistics
    ram_st_sst_mosi            : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_st_sst_miso            : OUT t_mem_miso := c_mem_miso_rst;

    -- Spectral Inversion
    reg_si_mosi                : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_si_miso                : OUT t_mem_miso := c_mem_miso_rst;

    -- Filter coefficients
    ram_fil_coefs_mosi         : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_fil_coefs_miso         : OUT t_mem_miso := c_mem_miso_rst;

    -- Equalizer gains
    ram_equalizer_gains_mosi   : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_equalizer_gains_miso   : OUT t_mem_miso := c_mem_miso_rst;

    -- DP Selector
    reg_dp_selector_mosi       : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_selector_miso       : OUT t_mem_miso := c_mem_miso_rst;

    ----------------------------------------------
    -- SDP Info 
    ----------------------------------------------
    reg_sdp_info_mosi          : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_sdp_info_miso          : OUT t_mem_miso := c_mem_miso_rst;

    ----------------------------------------------
    -- XSUB 
    ----------------------------------------------
    -- dp_sync_insert_v2
    reg_dp_sync_insert_v2_mosi  : IN  t_mem_mosi := c_mem_mosi_rst; 
    reg_dp_sync_insert_v2_miso  : OUT t_mem_miso := c_mem_miso_rst; 

    -- crosslets_info
    reg_crosslets_info_mosi     : IN  t_mem_mosi := c_mem_mosi_rst; 
    reg_crosslets_info_miso     : OUT t_mem_miso := c_mem_miso_rst; 

    -- bsn_scheduler_xsub
    reg_bsn_scheduler_xsub_mosi : IN  t_mem_mosi := c_mem_mosi_rst; 
    reg_bsn_scheduler_xsub_miso : OUT t_mem_miso := c_mem_miso_rst; 

    -- st_xsq
    ram_st_xsq_mosi             : IN  t_mem_mosi := c_mem_mosi_rst; 
    ram_st_xsq_miso             : OUT t_mem_miso := c_mem_miso_rst; 

    ----------------------------------------------
    -- BF 
    ----------------------------------------------
    -- Beamlet Subband Select
    ram_ss_ss_wide_mosi        : IN  t_mem_mosi := c_mem_mosi_rst;       
    ram_ss_ss_wide_miso        : OUT t_mem_miso := c_mem_miso_rst;

    -- Local BF bf weights
    ram_bf_weights_mosi        : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_bf_weights_miso        : OUT t_mem_miso := c_mem_miso_rst;

    -- mms_dp_scale Scale Beamlets
    reg_bf_scale_mosi          : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_bf_scale_miso          : OUT t_mem_miso := c_mem_miso_rst;

    -- Beamlet Data Output header fields
    reg_hdr_dat_mosi           : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_hdr_dat_miso           : OUT t_mem_miso := c_mem_miso_rst;

    -- Beamlet Data Output xonoff
    reg_dp_xonoff_mosi         : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_xonoff_miso         : OUT t_mem_miso := c_mem_miso_rst;

    -- Beamlet Statistics (BST)
    ram_st_bst_mosi            : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_st_bst_miso            : OUT t_mem_miso := c_mem_miso_rst;

    ----------------------------------------------
    -- SST 
    ----------------------------------------------
    -- Statistics Enable
    reg_stat_enable_sst_mosi       : IN  t_mem_mosi;
    reg_stat_enable_sst_miso       : OUT t_mem_miso;
    
    -- Statistics header info  
    reg_stat_hdr_dat_sst_mosi      : IN  t_mem_mosi;
    reg_stat_hdr_dat_sst_miso      : OUT t_mem_miso;

    ----------------------------------------------
    -- XST 
    ----------------------------------------------
    -- Statistics Enable
    reg_stat_enable_xst_mosi       : IN  t_mem_mosi;
    reg_stat_enable_xst_miso       : OUT t_mem_miso;
    
    -- Statistics header info  
    reg_stat_hdr_dat_xst_mosi      : IN  t_mem_mosi;
    reg_stat_hdr_dat_xst_miso      : OUT t_mem_miso;

    ----------------------------------------------
    -- BST 
    ----------------------------------------------
    -- Statistics Enable
    reg_stat_enable_bst_mosi      : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_stat_enable_bst_miso      : OUT t_mem_miso := c_mem_miso_rst;
    
    -- Statistics header info 
    reg_stat_hdr_dat_bst_mosi     : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_stat_hdr_dat_bst_miso     : OUT t_mem_miso := c_mem_miso_rst;

    -- QSFP serial
    unb2_board_front_io_serial_tx_arr : OUT STD_LOGIC_VECTOR(6 * c_quad-1 DOWNTO 0) := (OTHERS => '0');
    unb2_board_front_io_serial_rx_arr : IN  STD_LOGIC_VECTOR(6 * c_quad-1 DOWNTO 0) := (OTHERS => '0');

    -- QSFP LEDS
    unb2_board_qsfp_leds_tx_sosi_arr : OUT t_dp_sosi_arr(6 * c_quad-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
    unb2_board_qsfp_leds_tx_siso_arr : OUT t_dp_siso_arr(6 * c_quad-1 DOWNTO 0) := (OTHERS => c_dp_siso_rst);
    unb2_board_qsfp_leds_rx_sosi_arr : OUT t_dp_sosi_arr(6 * c_quad-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst)
  );
END sdp_station;


ARCHITECTURE str OF sdp_station IS

  -- 10 GbE Interface
  CONSTANT c_nof_10GbE_offload_streams     : NATURAL := 1;
  CONSTANT c_nof_blocks_per_packet         : NATURAL := 4;
  CONSTANT c_nof_beamlets_per_block        : NATURAL := c_sdp_N_pol * c_sdp_S_sub_bf; 
  CONSTANT c_10GbE_block_size              : NATURAL := c_nof_blocks_per_packet * c_nof_beamlets_per_block / 4; -- 4 beamlets fit in 1 64bit longword 
  CONSTANT c_fifo_tx_fill                  : NATURAL := c_10GbE_block_size; 
  CONSTANT c_fifo_tx_size                  : NATURAL := c_fifo_tx_fill + 11; -- Make fifo size large enough for adding header.
 
  -- Address widths of a single MM instance
  CONSTANT c_addr_w_ram_ss_ss_wide : NATURAL := ceil_log2(c_sdp_P_pfb * c_sdp_S_sub_bf * c_sdp_Q_fft); 
  CONSTANT c_addr_w_ram_bf_weights : NATURAL := ceil_log2(c_sdp_N_pol * c_sdp_P_pfb * c_sdp_S_sub_bf * c_sdp_Q_fft); 
  CONSTANT c_addr_w_reg_bf_scale   : NATURAL := 1; 
  CONSTANT c_addr_w_reg_hdr_dat    : NATURAL := ceil_log2(field_nof_words(c_sdp_cep_hdr_field_arr, c_word_w)); 
  CONSTANT c_addr_w_reg_dp_xonoff  : NATURAL := 1; 
  CONSTANT c_addr_w_ram_st_bst     : NATURAL := ceil_log2(c_sdp_S_sub_bf*c_sdp_N_pol*(c_longword_sz/c_word_sz));
  
  -- Read only sdp_info values
  CONSTANT c_f_adc     : STD_LOGIC := '1'; -- '0' => 160M, '1' => 200M
  CONSTANT c_fsub_type : STD_LOGIC := '0'; -- '0' => critical sampled PFB, '1' => oversampled PFB
  
  SIGNAL gn_index : NATURAL := 0;

  ----------------------------------------------
  -- BF 
  ----------------------------------------------
  -- Beamlet Subband Select
  SIGNAL ram_ss_ss_wide_mosi_arr    : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);       
  SIGNAL ram_ss_ss_wide_miso_arr    : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  -- Local BF bf weights
  SIGNAL ram_bf_weights_mosi_arr    : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL ram_bf_weights_miso_arr    : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  -- mms_dp_scale Scale Beamlets
  SIGNAL reg_bf_scale_mosi_arr      : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL reg_bf_scale_miso_arr      : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  -- Beamlet Data Output header fields
  SIGNAL reg_hdr_dat_mosi_arr       : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL reg_hdr_dat_miso_arr       : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  -- Beamlet Data Output xonoff
  SIGNAL reg_dp_xonoff_mosi_arr     : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL reg_dp_xonoff_miso_arr     : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  -- Beamlet Statistics (BST)
  SIGNAL ram_st_bst_mosi_arr        : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL ram_st_bst_miso_arr        : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  ----------------------------------------------
  -- BST 
  ----------------------------------------------
  -- Statistics Enable
  SIGNAL reg_stat_enable_bst_mosi_arr  : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL reg_stat_enable_bst_miso_arr  : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);
  
  -- Statistics header info 
  SIGNAL reg_stat_hdr_dat_bst_mosi_arr : t_mem_mosi_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL reg_stat_hdr_dat_bst_miso_arr : t_mem_miso_arr(c_sdp_N_beamsets-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);
  ----------------------------------------------

  SIGNAL ait_sosi_arr                      : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0);         
  SIGNAL pfb_sosi_arr                      : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0);         
  SIGNAL fsub_sosi_arr                     : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0);        
  
  SIGNAL dp_bsn_source_restart             : STD_LOGIC;
 
  SIGNAL bf_udp_sosi_arr                   : t_dp_sosi_arr(c_sdp_N_beamsets-1 DOWNTO 0);         
  SIGNAL bf_udp_siso_arr                   : t_dp_siso_arr(c_sdp_N_beamsets-1 DOWNTO 0);    
  SIGNAL bf_10GbE_hdr_fields_out_arr       : t_slv_1024_arr(c_sdp_N_beamsets-1 DOWNTO 0);

  -- 10GbE
  SIGNAL tr_ref_clk_312                    : STD_LOGIC;
  SIGNAL tr_ref_clk_156                    : STD_LOGIC;
  SIGNAL tr_ref_rst_156                    : STD_LOGIC;


  SIGNAL nw_10gbe_snk_in_arr               : t_dp_sosi_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL nw_10gbe_snk_out_arr              : t_dp_siso_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  SIGNAL nw_10gbe_src_out_arr              : t_dp_sosi_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL nw_10gbe_src_in_arr               : t_dp_siso_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);

  SIGNAL nw_10GbE_hdr_fields_in_arr        : t_slv_1024_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0);


  SIGNAL cep_eth_src_mac                   : STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0);
  SIGNAL cep_ip_src_addr                   : STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0);
  SIGNAL cep_udp_src_port                  : STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);
  SIGNAL stat_eth_src_mac                  : STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0);
  SIGNAL stat_ip_src_addr                  : STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0);
  SIGNAL sst_udp_src_port                  : STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);
  SIGNAL bst_udp_src_port                  : STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);
  SIGNAL xst_udp_src_port                  : STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);

  SIGNAL sdp_info                          : t_sdp_info := c_sdp_info_rst;

BEGIN

  -----------------------------------------------------------------------------
  -- SDP Info register
  -----------------------------------------------------------------------------
  gn_index <= TO_UINT(gn_id);
  -- derive MAC, IP and UDP Port 
  cep_eth_src_mac  <= c_sdp_cep_eth_src_mac_47_16 & RESIZE_UVEC(this_bck_id, c_byte_w) & RESIZE_UVEC(this_chip_id, c_byte_w); -- Simply use chip_id since we only use 1 of the 6*4 = 24 10GbE port.
  cep_ip_src_addr  <= c_sdp_cep_ip_src_addr_31_16 & RESIZE_UVEC(this_bck_id, c_byte_w) & INCR_UVEC(RESIZE_UVEC(this_chip_id, c_byte_w), 1); -- +1 to avoid IP = *.*.*.0
  cep_udp_src_port <= c_sdp_cep_udp_src_port_15_8 & RESIZE_UVEC(gn_id, c_byte_w);
  
  stat_eth_src_mac <= c_sdp_stat_eth_src_mac_47_16 & RESIZE_UVEC(this_bck_id, c_byte_w) & RESIZE_UVEC(this_chip_id, c_byte_w); -- Simply use chip_id since we only use 1 of the 6*4 = 24 10GbE port.
  stat_ip_src_addr <= c_sdp_stat_ip_src_addr_31_16 & RESIZE_UVEC(this_bck_id, c_byte_w) & INCR_UVEC(RESIZE_UVEC(this_chip_id, c_byte_w), 1); -- +1 to avoid IP = *.*.*.0
  sst_udp_src_port <= c_sdp_sst_udp_src_port_15_8 & RESIZE_UVEC(gn_id, c_byte_w);
  bst_udp_src_port <= c_sdp_bst_udp_src_port_15_8 & RESIZE_UVEC(gn_id, c_byte_w);
  xst_udp_src_port <= c_sdp_xst_udp_src_port_15_8 & RESIZE_UVEC(gn_id, c_byte_w);

  u_sdp_info : ENTITY work.sdp_info
  PORT MAP(
    -- Clocks and reset
    mm_rst    => mm_rst,  -- reset synchronous with mm_clk
    mm_clk    => mm_clk,  -- memory-mapped bus clock

    dp_clk    => dp_clk,
    dp_rst    => dp_rst,

    reg_mosi  => reg_sdp_info_mosi,
    reg_miso  => reg_sdp_info_miso,

    -- inputs from other blocks
    gn_index  => gn_index, 
    f_adc     => c_f_adc, 
    fsub_type => c_fsub_type, 

    -- sdp info
    sdp_info => sdp_info 
  ); 

  -----------------------------------------------------------------------------
  -- node_adc_input_and_timing (AIT)
  --   .Contains JESD receiver, bsn source and associated data buffers, diagnostics and statistics
  -----------------------------------------------------------------------------
  u_ait: ENTITY work.node_sdp_adc_input_and_timing
  GENERIC MAP(
    g_technology                => g_technology,
    g_sim                       => g_sim,
    g_bsn_nof_clk_per_sync      => g_bsn_nof_clk_per_sync                
  )
  PORT MAP(
    -- clocks and resets
    mm_clk                      => mm_clk,           
    mm_rst                      => mm_rst,           
    dp_clk                      => dp_clk,           
    dp_rst                      => dp_rst,           
 
    -- mm control buses
    jesd_ctrl_mosi              => jesd_ctrl_mosi, 
    jesd_ctrl_miso              => jesd_ctrl_miso, 
    jesd204b_mosi               => jesd204b_mosi,         
    jesd204b_miso               => jesd204b_miso,         
    reg_dp_shiftram_mosi        => reg_dp_shiftram_mosi,
    reg_dp_shiftram_miso        => reg_dp_shiftram_miso,
    reg_bsn_source_v2_mosi      => reg_bsn_source_v2_mosi,
    reg_bsn_source_v2_miso      => reg_bsn_source_v2_miso,
    reg_bsn_scheduler_wg_mosi   => reg_bsn_scheduler_wg_mosi,
    reg_bsn_scheduler_wg_miso   => reg_bsn_scheduler_wg_miso,
    reg_wg_mosi                 => reg_wg_mosi,
    reg_wg_miso                 => reg_wg_miso,
    ram_wg_mosi                 => ram_wg_mosi,
    ram_wg_miso                 => ram_wg_miso,
    reg_bsn_monitor_input_mosi  => reg_bsn_monitor_input_mosi,
    reg_bsn_monitor_input_miso  => reg_bsn_monitor_input_miso,
    ram_diag_data_buf_bsn_mosi  => ram_diag_data_buf_bsn_mosi,
    ram_diag_data_buf_bsn_miso  => ram_diag_data_buf_bsn_miso,
    reg_diag_data_buf_bsn_mosi  => reg_diag_data_buf_bsn_mosi,
    reg_diag_data_buf_bsn_miso  => reg_diag_data_buf_bsn_miso,
    reg_aduh_monitor_mosi       => reg_aduh_monitor_mosi,
    reg_aduh_monitor_miso       => reg_aduh_monitor_miso,
  
     -- Jesd external IOs
    jesd204b_serial_data       => JESD204B_SERIAL_DATA,
    jesd204b_refclk            => JESD204B_REFCLK,   
    jesd204b_sysref            => JESD204B_SYSREF,   
    jesd204b_sync_n            => JESD204B_SYNC_N,   
 
    -- Streaming data output
    out_sosi_arr               => ait_sosi_arr,
    dp_bsn_source_restart      => dp_bsn_source_restart        
  );

  -----------------------------------------------------------------------------
  -- node_sdp_filterbank (FSUB)
  -----------------------------------------------------------------------------
  gen_use_fsub : IF g_use_fsub GENERATE
    u_fsub : ENTITY work.node_sdp_filterbank 
    GENERIC MAP(
      g_sim                    => g_sim,
      g_wpfb                   => g_wpfb,
      g_scope_selected_subband => g_scope_selected_subband
    )
    PORT MAP(
      dp_clk                => dp_clk, 
      dp_rst                => dp_rst, 
                                                 
      in_sosi_arr           => ait_sosi_arr,    
      pfb_sosi_arr          => pfb_sosi_arr,
      fsub_sosi_arr         => fsub_sosi_arr,
      dp_bsn_source_restart => dp_bsn_source_restart,

      sst_udp_sosi          => udp_tx_sosi_arr(0),
      sst_udp_siso          => udp_tx_siso_arr(0),
                                                 
      mm_rst                => mm_rst, 
      mm_clk                => mm_clk, 
                                                 
      reg_si_mosi           => reg_si_mosi, 
      reg_si_miso           => reg_si_miso, 
      ram_st_sst_mosi       => ram_st_sst_mosi,  
      ram_st_sst_miso       => ram_st_sst_miso, 
      ram_fil_coefs_mosi    => ram_fil_coefs_mosi,  
      ram_fil_coefs_miso    => ram_fil_coefs_miso,
      ram_gains_mosi        => ram_equalizer_gains_mosi,     
      ram_gains_miso        => ram_equalizer_gains_miso,     
      reg_selector_mosi     => reg_dp_selector_mosi,  
      reg_selector_miso     => reg_dp_selector_miso,

      reg_enable_mosi       => reg_stat_enable_sst_mosi,
      reg_enable_miso       => reg_stat_enable_sst_miso,
      reg_hdr_dat_mosi      => reg_stat_hdr_dat_sst_mosi,
      reg_hdr_dat_miso      => reg_stat_hdr_dat_sst_miso,
  
      sdp_info              => sdp_info,
      gn_id                 => gn_id,
      eth_src_mac           => stat_eth_src_mac,
      ip_src_addr           => stat_ip_src_addr,
      udp_src_port          => sst_udp_src_port
    );
  END GENERATE;


  -----------------------------------------------------------------------------
  -- node_sdp_correlator (XSUB)
  -----------------------------------------------------------------------------
  gen_use_xsub : IF g_use_xsub GENERATE
    u_xsub : ENTITY work.node_sdp_correlator 
    GENERIC MAP(
      g_sim  => g_sim,
      g_P_sq => g_P_sq
    )
    PORT MAP(
      dp_clk                      => dp_clk, 
      dp_rst                      => dp_rst, 
                                                       
      in_sosi_arr                 => fsub_sosi_arr,    

      xst_udp_sosi                => udp_tx_sosi_arr(1),
      xst_udp_siso                => udp_tx_siso_arr(1), 
                                                      
      mm_rst                      => mm_rst, 
      mm_clk                      => mm_clk, 
                                              
      reg_dp_sync_insert_v2_mosi  => reg_dp_sync_insert_v2_mosi, 
      reg_dp_sync_insert_v2_miso  => reg_dp_sync_insert_v2_miso,  
      reg_crosslets_info_mosi     => reg_crosslets_info_mosi,     
      reg_crosslets_info_miso     => reg_crosslets_info_miso,     
      reg_bsn_scheduler_xsub_mosi => reg_bsn_scheduler_xsub_mosi, 
      reg_bsn_scheduler_xsub_miso => reg_bsn_scheduler_xsub_miso, 
      ram_st_xsq_mosi             => ram_st_xsq_mosi,             
      ram_st_xsq_miso             => ram_st_xsq_miso,

      reg_stat_enable_mosi        => reg_stat_enable_xst_mosi,
      reg_stat_enable_miso        => reg_stat_enable_xst_miso,
      reg_stat_hdr_dat_mosi       => reg_stat_hdr_dat_xst_mosi,
      reg_stat_hdr_dat_miso       => reg_stat_hdr_dat_xst_miso,
  
      sdp_info                    => sdp_info,
      gn_id                       => gn_id,
      stat_eth_src_mac            => stat_eth_src_mac,
      stat_ip_src_addr            => stat_ip_src_addr,
      stat_udp_src_port           => xst_udp_src_port
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- nof beamsets node_sdp_beamformers (BF)
  -----------------------------------------------------------------------------
  gen_use_bf : IF g_use_bf GENERATE
    -- Beamformers
    gen_bf : FOR beamset_id IN 0 TO c_sdp_N_beamsets-1 GENERATE
      u_bf : ENTITY work.node_sdp_beamformer
      GENERIC MAP(
        g_sim                    => g_sim, 
        g_beamset_id             => beamset_id, 
        g_scope_selected_beamlet => g_scope_selected_subband 
      )
      PORT MAP(
        dp_clk                   => dp_clk,  
        dp_rst                   => dp_rst,  
      
        in_sosi_arr              => fsub_sosi_arr, 
        bf_udp_sosi              => bf_udp_sosi_arr(beamset_id),
        bf_udp_siso              => bf_udp_siso_arr(beamset_id),
        bst_udp_sosi             => udp_tx_sosi_arr(2+ beamset_id),  
        bst_udp_siso             => udp_tx_siso_arr(2+ beamset_id),  
      
        mm_rst                   => mm_rst,  
        mm_clk                   => mm_clk,  
      
        ram_ss_ss_wide_mosi      => ram_ss_ss_wide_mosi_arr(beamset_id),  
        ram_ss_ss_wide_miso      => ram_ss_ss_wide_miso_arr(beamset_id), 
        ram_bf_weights_mosi      => ram_bf_weights_mosi_arr(beamset_id), 
        ram_bf_weights_miso      => ram_bf_weights_miso_arr(beamset_id), 
        reg_bf_scale_mosi        => reg_bf_scale_mosi_arr(beamset_id), 
        reg_bf_scale_miso        => reg_bf_scale_miso_arr(beamset_id), 
        reg_hdr_dat_mosi         => reg_hdr_dat_mosi_arr(beamset_id), 
        reg_hdr_dat_miso         => reg_hdr_dat_miso_arr(beamset_id), 
        reg_dp_xonoff_mosi       => reg_dp_xonoff_mosi_arr(beamset_id), 
        reg_dp_xonoff_miso       => reg_dp_xonoff_miso_arr(beamset_id), 
        ram_st_bst_mosi          => ram_st_bst_mosi_arr(beamset_id), 
        ram_st_bst_miso          => ram_st_bst_miso_arr(beamset_id), 
        reg_stat_enable_mosi     => reg_stat_enable_bst_mosi_arr(beamset_id),
        reg_stat_enable_miso     => reg_stat_enable_bst_miso_arr(beamset_id),
        reg_stat_hdr_dat_mosi    => reg_stat_hdr_dat_bst_mosi_arr(beamset_id),
        reg_stat_hdr_dat_miso    => reg_stat_hdr_dat_bst_miso_arr(beamset_id),     
 
        sdp_info                 => sdp_info,
        gn_id                    => gn_id,
        eth_src_mac              => cep_eth_src_mac, 
        ip_src_addr              => cep_ip_src_addr, 
        udp_src_port             => cep_udp_src_port, 
        stat_eth_src_mac         => stat_eth_src_mac, 
        stat_ip_src_addr         => stat_ip_src_addr, 
        stat_udp_src_port        => bst_udp_src_port, 
            
        hdr_fields_out           => bf_10GbE_hdr_fields_out_arr(beamset_id)
      );
  
    END GENERATE;
  
    -- MM multiplexing
    u_mem_mux_ram_ss_ss_wide : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_addr_w_ram_ss_ss_wide
    )
    PORT MAP (
      mosi     => ram_ss_ss_wide_mosi,
      miso     => ram_ss_ss_wide_miso,
      mosi_arr => ram_ss_ss_wide_mosi_arr,
      miso_arr => ram_ss_ss_wide_miso_arr
    );
  
    u_mem_mux_ram_bf_weights : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_addr_w_ram_bf_weights
    )
    PORT MAP (
      mosi     => ram_bf_weights_mosi,
      miso     => ram_bf_weights_miso,
      mosi_arr => ram_bf_weights_mosi_arr,
      miso_arr => ram_bf_weights_miso_arr
    );
  
    u_mem_mux_reg_bf_scale : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_addr_w_reg_bf_scale
    )
    PORT MAP (
      mosi     => reg_bf_scale_mosi,
      miso     => reg_bf_scale_miso,
      mosi_arr => reg_bf_scale_mosi_arr,
      miso_arr => reg_bf_scale_miso_arr
    );
  
    u_mem_mux_reg_hdr_dat : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_addr_w_reg_hdr_dat
    )
    PORT MAP (
      mosi     => reg_hdr_dat_mosi,
      miso     => reg_hdr_dat_miso,
      mosi_arr => reg_hdr_dat_mosi_arr,
      miso_arr => reg_hdr_dat_miso_arr
    );
  
    u_mem_mux_reg_dp_xonoff : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_addr_w_reg_dp_xonoff
    )
    PORT MAP (
      mosi     => reg_dp_xonoff_mosi,
      miso     => reg_dp_xonoff_miso,
      mosi_arr => reg_dp_xonoff_mosi_arr,
      miso_arr => reg_dp_xonoff_miso_arr
    );
  
    u_mem_mux_ram_st_bst : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_addr_w_ram_st_bst
    )
    PORT MAP (
      mosi     => ram_st_bst_mosi,
      miso     => ram_st_bst_miso,
      mosi_arr => ram_st_bst_mosi_arr,
      miso_arr => ram_st_bst_miso_arr
    );

    u_mem_mux_reg_stat_enable_bst : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_sdp_reg_stat_enable_addr_w
    )
    PORT MAP (
      mosi     => reg_stat_enable_bst_mosi,
      miso     => reg_stat_enable_bst_miso,
      mosi_arr => reg_stat_enable_bst_mosi_arr,
      miso_arr => reg_stat_enable_bst_miso_arr
    );
 
    u_mem_mux_reg_stat_hdr_dat_bst : ENTITY common_lib.common_mem_mux
    GENERIC MAP (
      g_nof_mosi    => c_sdp_N_beamsets,
      g_mult_addr_w => c_sdp_reg_stat_hdr_dat_addr_w
    )
    PORT MAP (
      mosi     => reg_stat_hdr_dat_bst_mosi,
      miso     => reg_stat_hdr_dat_bst_miso,
      mosi_arr => reg_stat_hdr_dat_bst_mosi_arr,
      miso_arr => reg_stat_hdr_dat_bst_miso_arr
    );
   
    -----------------------------------------------------------------------------
    -- DP MUX
    -----------------------------------------------------------------------------
    -- Assign hdr_fields to nw_10GbE for ARP/PING functionality. Only the fields: 
    -- eth_src_mac, ip_src_addr and ip_dst_addr are used. Which are identical for
    -- both beamsets.
    nw_10GbE_hdr_fields_in_arr(0) <= bf_10GbE_hdr_fields_out_arr(0);
  
    u_dp_mux : ENTITY dp_lib.dp_mux
    GENERIC MAP (
      g_nof_input => c_sdp_N_beamsets,
      g_sel_ctrl_invert => TRUE,
      g_fifo_size => array_init(0,c_sdp_N_beamsets), --no FIFO used but must match g_nof_input
      g_fifo_fill => array_init(0,c_sdp_N_beamsets) --no FIFO used but must match g_nof_input
    )
    PORT MAP (
      clk => dp_clk,
      rst => dp_rst,
  
      snk_in_arr  => bf_udp_sosi_arr,
      snk_out_arr => bf_udp_siso_arr,
    
      src_out => nw_10gbe_snk_in_arr(0),
      src_in  => nw_10gbe_snk_out_arr(0)
    );

    ---------------
    -- nw_10GbE
    ---------------
    u_nw_10GbE: ENTITY nw_10GbE_lib.nw_10GbE
    GENERIC MAP (
      g_technology    => g_technology,
      g_sim           => g_sim,
      g_sim_level     => 1,
      g_nof_macs      => c_nof_10GbE_offload_streams,
      g_direction     => "TX_RX",
      g_tx_fifo_fill  => c_fifo_tx_fill,
      g_tx_fifo_size  => c_fifo_tx_size,
      g_ip_hdr_field_arr => c_sdp_cep_hdr_field_arr
  
    )
    PORT MAP (
      -- Transceiver PLL reference clock
      tr_ref_clk_644        => SA_CLK,
      tr_ref_clk_312        => tr_ref_clk_312,  
      tr_ref_clk_156        => tr_ref_clk_156,  
      tr_ref_rst_156        => tr_ref_rst_156,  
  
      -- MM interface
      mm_rst                => mm_rst,
      mm_clk                => mm_clk,
  
      reg_mac_mosi          => reg_nw_10GbE_mac_mosi,
      reg_mac_miso          => reg_nw_10GbE_mac_miso,
  
      reg_eth10g_mosi       => reg_nw_10GbE_eth10g_mosi,
      reg_eth10g_miso       => reg_nw_10GbE_eth10g_miso,
  
      -- DP interface
      dp_rst                => dp_rst,
      dp_clk                => dp_clk,
      dp_pps                => dp_pps,
  
      src_out_arr           => nw_10gbe_src_out_arr,
      src_in_arr            => nw_10gbe_src_in_arr,
  
      snk_out_arr           => nw_10gbe_snk_out_arr,
      snk_in_arr            => nw_10gbe_snk_in_arr,
  
      -- Serial IO
      serial_tx_arr         => unb2_board_front_io_serial_tx_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0), 
      serial_rx_arr         => unb2_board_front_io_serial_rx_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0),
  
      hdr_fields_in_arr     => nw_10GbE_hdr_fields_in_arr 
    );
  END GENERATE;
  
  ---------
  -- PLL
  ---------
  u_tech_pll_xgmii_mac_clocks : ENTITY tech_pll_lib.tech_pll_xgmii_mac_clocks
  GENERIC MAP (
    g_technology => g_technology
  )
  PORT MAP (
    refclk_644 => SA_CLK,
    rst_in     => mm_rst,
    clk_156    => tr_ref_clk_156,
    clk_312    => tr_ref_clk_312,
    rst_156    => tr_ref_rst_156,
    rst_312    => OPEN
  );
  
  ------------
  -- LEDs
  ------------
  unb2_board_qsfp_leds_tx_siso_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) <= nw_10gbe_snk_out_arr;
  unb2_board_qsfp_leds_tx_sosi_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) <= nw_10gbe_snk_in_arr;
  unb2_board_qsfp_leds_rx_sosi_arr(c_nof_10GbE_offload_streams-1 DOWNTO 0) <= nw_10gbe_src_out_arr;


END str;
